// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/24DSI64C200K/24DSI64C200K_linux_1.x_gsc_dn/driver/device.c $
// $Rev: 52095 $
// $Date: 2022-12-01 10:53:33 -0600 (Thu, 01 Dec 2022) $

// 24DSI64C200K: Device Driver: source file

#include "main.h"



// variables ******************************************************************

const gsc_dev_id_t	dev_id_list[]	=
{
	// model			Vendor	Device	SubVen	SubDev	type
	{ "24DSI64C200K",	0x10B5, 0x9056, 0x10B5, 0x3575,	GSC_DEV_TYPE_24DSI64C200K	},
	{ NULL }
};



//*****************************************************************************
static void _channels_compute(dev_data_t* dev)
{
	s32	v	= GSC_FIELD_DECODE(dev->cache.gsc_bcfgr_32, 17, 16);

	dev->cache.channels_max	= 64;

	switch (v)
	{
		default:

			printf(	"%s, #%d: Board Config Register 0x%lX, D16-D18 unrecognized.\n",
					DEV_NAME,
					dev->board_index,
					(long) dev->cache.gsc_bcfgr_32);

			FALLTHROUGH;	/* fall through */

		case 2:	dev->cache.channel_qty	= 32;	break;
		case 1:	dev->cache.channel_qty	= 48;	break;
		case 0:	dev->cache.channel_qty	= 64;	break;
	}
}



//*****************************************************************************
static void _filter_freq_compute(dev_data_t* dev)
{
	s32	v	= GSC_FIELD_DECODE(dev->cache.gsc_bcfgr_32, 21, 20);

	switch (v)
	{
		default:

			printf(	"%s, #%d: Board Config Register 0x%lX, D20-D21 unrecognized.\n",
					DEV_NAME,
					dev->board_index,
					(long) dev->cache.gsc_bcfgr_32);
			dev->cache.filter_freq	= -1;
			break;

		case 0:	dev->cache.filter_freq	= 900000;	break;
	}
}



//*****************************************************************************
static void _master_clock_compute(dev_data_t* dev)
{
	s32	v	= GSC_FIELD_DECODE(dev->cache.gsc_bcfgr_32, 23, 22);

	switch (v)
	{
		default:

			printf(	"%s, #%d: Board Config Register 0x%lX, D22-D23 unrecognized.\n",
					DEV_NAME,
					dev->board_index,
					(long) dev->cache.gsc_bcfgr_32);

			FALLTHROUGH;	/* fall through */

		case 0:	dev->cache.master_clock	= 49152000;	break;
	}
}



//*****************************************************************************
static void _voltage_range_compute(dev_data_t* dev)
{
	s32	v	= GSC_FIELD_DECODE(dev->cache.gsc_bcfgr_32, 19, 18);

	switch (v)
	{
		default:

			printf(	"%s, #%d: Board Config Register 0x%lX, D18-D19 unrecognized.\n",
					DEV_NAME,
					dev->board_index,
					(long) dev->cache.gsc_bcfgr_32);
			dev->cache.filter_freq	= -1;
			break;

		case 0:	dev->cache.voltage_range	= DSI64C200K_QUERY_V_RANGE_10;	break;
	}
}



/******************************************************************************
*
*	Function:	dev_device_create
*
*	Purpose:
*
*		Do everything needed to setup and use the given device.
*
*	Arguments:
*
*		dev		The structure to initialize.
*
*	Returned:
*
*		0		All is well.
*		< 0		An appropriate error status.
*
******************************************************************************/

int dev_device_create(dev_data_t* dev)
{
	static const gsc_bar_maps_t	bar_map	=
	{
		{
			// mem	io	rw
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 0: PLX registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 1: PLX registers, I/O mapped
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RW },	// BAR 2: GSC registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 3: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 4: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO }	// BAR 5: unused
		}
	};

	u32	dma;
	int	ret;

	for (;;)	// A convenience loop.
	{
		// Verify some macro contents.
		ret	= gsc_macro_test_base_name(DSI64C200K_BASE_NAME);
		if (ret)	break;

		ret	= gsc_macro_test_model();
		if (ret)	break;

		// PCI setup.
		ret	= os_pci_dev_enable(&dev->pci);
		if (ret)	break;

		ret	= os_pci_master_set(&dev->pci);
		if (ret)	break;

		// Control ISR access to the device and data structure.
		ret	= os_spinlock_create(&dev->spinlock);
		if (ret)	break;

		// Control access to the device and data structure.
		ret	= os_sem_create(&dev->sem);
		if (ret)	break;

		// Access the BAR regions.
		ret	= gsc_bar_create(dev, &dev->bar, &bar_map);
		if (ret)	break;

		// Firmware access.
		dev->vaddr.gsc_bctlr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI64C200K_GSC_BCTLR));
		dev->vaddr.gsc_rcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI64C200K_GSC_RCR));
		dev->vaddr.gsc_diopr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI64C200K_GSC_DIOPR));
		dev->vaddr.gsc_aicr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI64C200K_GSC_AICR));
		dev->vaddr.gsc_bbsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI64C200K_GSC_BBSR));
		dev->vaddr.gsc_bttr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI64C200K_GSC_BTTR));
		dev->vaddr.gsc_bufcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI64C200K_GSC_BUFCR));
		dev->vaddr.gsc_bcfgr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI64C200K_GSC_BCFGR));
		dev->vaddr.gsc_bufsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI64C200K_GSC_BUFSR));
		dev->vaddr.gsc_avr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI64C200K_GSC_AVR));
		dev->vaddr.gsc_idbr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(DSI64C200K_GSC_IDBR));
		dev->vaddr.gsc_asiocr_32	= GSC_VADDR(dev, GSC_REG_OFFSET(DSI64C200K_GSC_ASIOCR));

		// Data cache initialization.
		dev->cache.gsc_bcfgr_32		= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcfgr_32);

		dev->model					= dev_id_list[0].model;
		dev->cache.auto_cal_ms		= 25000;
		dev->cache.fgen_max			= 68000000L;
		dev->cache.fgen_min			= 34000000L;
		dev->cache.fifo_size		= _256K;
		dev->cache.fref_default		= 32768000L;
		dev->cache.fsamp_max		= 250000;
		dev->cache.fsamp_min		= 33;
		dev->cache.fw_rev			= GSC_FIELD_DECODE(dev->cache.gsc_bcfgr_32, 11, 0);
		dev->cache.init_adc_ms		= 200;
		dev->cache.initialize_ms	= 400;
		dev->cache.mclk_max			= 68000000L;
		dev->cache.mclk_min			= 34000000L;
		dev->cache.nref_max			= 300;
		dev->cache.nref_min			= 20;
		dev->cache.nref_opt_max		= 100;
		dev->cache.nref_opt_min		= 25;
		dev->cache.nvco_max			= 300;
		dev->cache.nvco_min			= 20;
		dev->cache.nvco_opt_max		= 100;
		dev->cache.nvco_opt_min		= 25;

		_channels_compute(dev);
		_voltage_range_compute(dev);
		_filter_freq_compute(dev);
		_master_clock_compute(dev);

		// Initialize additional resources.
		ret	= dev_irq_create(dev);
		if (ret)	break;

		ret	= dev_io_create(dev);
		if (ret)	break;

		dma	= GSC_DMA_SEL_STATIC
			| GSC_DMA_CAP_BMDMA_READ
			| GSC_DMA_CAP_DMDMA_READ;
		ret	= gsc_dma_create(dev, dma, dma);

		if (ret == 0)
			os_metrics(dev, dev->vaddr.gsc_diopr_32);

		break;
	}

	return(ret);
}



/******************************************************************************
*
*	Function:	dev_device_destroy
*
*	Purpose:
*
*		Do everything needed to release the referenced device.
*
*	Arguments:
*
*		dev		The partial data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dev_device_destroy(dev_data_t* dev)
{
	if (dev)
	{
		gsc_dma_destroy(dev);
		dev_io_destroy(dev);
		dev_irq_destroy(dev);
		gsc_bar_destroy(&dev->bar);
		os_sem_destroy(&dev->sem);
		os_spinlock_destroy(&dev->spinlock);
		os_pci_master_clear(&dev->pci);
		os_pci_dev_disable(&dev->pci);
	}
}


