// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/24DSI64C200K/24DSI64C200K_linux_1.x_gsc_dn/samples/irq/perform.c $
// $Rev: 52122 $
// $Date: 2022-12-01 11:45:05 -0600 (Thu, 01 Dec 2022) $

// 24DSI64C200K: Sample Application: source file

#include "main.h"



// data types *****************************************************************

typedef struct
{
	const char*	name;
	s32			main;
	s32			gsc;
	s32			alt;
	s32			io;
	int			(*func)(int fd, int supported);
} irq_test_t;



// variables ******************************************************************

static	int			_fd;
static	os_sem_t	_sem_enter;
static	os_sem_t	_sem_exit;
static	int			_thread_done;
static	char		_thread_err[256];
static	gsc_wait_t	_thread_wait;

static const irq_test_t	_tests[]	=
{
	// Main interrupts
	// name						main										gsc	alt	io	func
	{ "PCI",					GSC_WAIT_MAIN_PCI,							0,	0,	0,	main_pci_irq_test				},
	{ "DMA0/DMA1",				GSC_WAIT_MAIN_DMA0 | GSC_WAIT_MAIN_DMA1,	0,	0,	0,	main_dma_irq_test				},
	{ "GSC",					GSC_WAIT_MAIN_GSC,							0,	0,	0,	main_gsc_irq_test				},

	// GSC Interrupts
	// name						main	gsc										alt	io	func
	{ "Initialize Done",		0,		DSI64C200K_WAIT_GSC_INIT_DONE,			0,	0,	gsc_init_done_irq_test			},
	{ "Auto-Calibration Done",	0,		DSI64C200K_WAIT_GSC_AUTO_CAL_DONE,		0,	0,	gsc_auto_cal_done_irq_test		},
	{ "Channels Ready Lo->Hi",	0,		DSI64C200K_WAIT_GSC_CHAN_READY_L2H,		0,	0,	gsc_chan_ready_l2h_irq_test		},
	{ "Channels Ready Hi->Lo",	0,		DSI64C200K_WAIT_GSC_CHAN_READY_H2L,		0,	0,	gsc_chan_ready_h2l_irq_test		},
	{ "Thresh Status Lo->Hi",	0,		DSI64C200K_WAIT_GSC_AI_BUF_THR_L2H,		0,	0,	gsc_thresh_status_l2h_irq_test	},
	{ "Thresh Status Hi->Lo",	0,		DSI64C200K_WAIT_GSC_AI_BUF_THR_H2L,		0,	0,	gsc_thresh_status_h2l_irq_test	},
	{ "AI Burst Done",			0,		DSI64C200K_WAIT_GSC_AI_BURST_DONE,		0,	0,	gsc_ai_burst_done_irq_test		},
	{ NULL }
};



//*****************************************************************************
static int _wait_thread(void* arg)
{
	int			errs;
	irq_test_t*	ptr		= (void*) arg;

	os_sem_unlock(&_sem_enter);

	memset(&_thread_wait, 0, sizeof(_thread_wait));
	_thread_wait.main	= ptr->main;
	_thread_wait.gsc	= ptr->gsc;
	_thread_wait.alt	= ptr->alt;
	_thread_wait.io		= ptr->io;
	_thread_wait.timeout_ms	= 15000;
	errs				= dsi64c200k_wait_event(_fd, -1, 0, &_thread_wait);

	if (errs)
		sprintf(_thread_err, "%d. Wait request failed.", __LINE__);

	os_sem_unlock(&_sem_exit);

	_thread_done	= 1;
	return(errs);
}



//*****************************************************************************
static int _wait_for_waiting(int fd, const irq_test_t* ptr)
{
	int			errs	= 0;
	int			i;
	int			limit	= 1000;
	gsc_wait_t	wait;

	memset(&wait, 0, sizeof(wait));
	wait.main	= ptr->main;
	wait.gsc	= ptr->gsc;
	wait.alt	= ptr->alt;
	wait.io		= ptr->io;

	for (i = 0;; i++)
	{
		errs	= dsi64c200k_wait_status(fd, -1, 0, &wait);

		if (errs)
		{
			printf("FAIL <---  (%d. Wait status request failed.)\n", __LINE__);
			break;
		}
		else if (i > limit)
		{
			printf("FAIL <---  (%d. Excessive delay for thread to wait.)\n", __LINE__);
			errs++;
			break;
		}
		else if (wait.count)
		{
			break;
		}
		else
		{
			os_sleep_ms(10);
		}
	}

	return(errs);
}



//*****************************************************************************
static int _wait_for_done(void)
{
	int	errs	= 0;
	int	i;

	for (i = 0;; i++)
	{
		if (_thread_done)
			break;

		if (i > 1000)
		{
			printf("FAIL <---  (%d. Excessive delay for thread to complete.)\n", __LINE__);
			errs++;
			break;
		}
		else
		{
			os_sleep_ms(10);
		}
	}

	return(errs);
}



//*****************************************************************************
static int _wait_cancel(int fd)
{
	int			errs	= 0;
	int			ret;
	gsc_wait_t	wait;

	memset(&wait, 0, sizeof(wait));
	wait.main	= 0xFFFFFFFF;
	wait.gsc	= 0xFFFFFFFF;
	wait.alt	= 0xFFFFFFFF;
	wait.io		= 0xFFFFFFFF;
	ret			= dsi64c200k_wait_cancel(fd, -1, 0, &wait);
	errs		+= ret ? 1 : 0;

	return(errs);
}



//*****************************************************************************
static int _test_irq(int fd, const irq_test_t* ptr)
{
	int			errs		= 0;
	int			ret;
	os_thread_t	thread;

	for (;;)	// A cenvenience loop.
	{
		gsc_label(ptr->name);

		// Prepare for testing the interrupt.
		_thread_err[0]	= 0;
		_thread_done	= 0;
		_fd				= fd;
		memset(&_thread_wait, 0, sizeof(_thread_wait));
		os_sem_create_qty(&_sem_enter, 1, 0);
		os_sem_create_qty(&_sem_exit, 1, 0);

		// Find out what level of support there is.
		ret	= ptr->func(fd, 1);

		if (ret > 0)
		{
			errs++;
			printf("FAIL <---  (%d. Request for support status failed.)\n", __LINE__);
			break;
		}

		if (ret == STATUS_NOT_SUPPORTED)
		{
			printf("SKIPPED  (%d. Interrupt not supported on this device.)\n", __LINE__);
			break;
		}

		if (ret == STATUS_NOT_IMPLEMENTED)
		{
			printf("SKIPPED  (%d. Interrupt test not implemented.)\n", __LINE__);
			break;
		}

		if (ret)
		{
			errs++;
			printf("SKIPPED  (%d. Invalid support status returned: %d)\n", __LINE__, ret);
			break;
		}

		// Create a thread to wait on the interrupt.
		errs	= os_thread_create(&thread, "wait thtread", _wait_thread, (void*) ptr);

		if (errs)
		{
			printf("FAIL <---  (%d. Thread creation failed.)\n", __LINE__);
			break;
		}

		// Wait for the thread to start running.
		ret	= os_sem_lock(&_sem_enter);

		if (ret)
		{
			errs++;
			printf("FAIL <---  (%d. Semaphore 'enter' lock failure.)\n", __LINE__);
			_wait_cancel(fd);
			os_sem_lock(&_sem_exit);
			_wait_for_done();
			os_thread_destroy(&thread);
			break;
		}

		// Wait for the thread to become "waiting".
		errs	= _wait_for_waiting(fd, ptr);

		if (errs)
		{
			_wait_cancel(fd);
			os_sem_lock(&_sem_exit);
			_wait_for_done();
			os_thread_destroy(&thread);
			break;
		}

		// Initiate the interrupt.
		errs			= ptr->func(fd, 0);

		if (errs)
		{
			printf("FAIL <---  (%d. Interrupt function failure.)\n", __LINE__);
			_wait_cancel(fd);
			os_sem_lock(&_sem_exit);
			_wait_for_done();
			os_thread_destroy(&thread);
			break;
		}

		// Wait for the thread to stop running.
		ret	= os_sem_lock(&_sem_exit);

		if (ret)
		{
			errs++;
			printf("FAIL <---  (%d. Semaphore 'exit' lock failure.)\n", __LINE__);
			_wait_cancel(fd);
			_wait_for_done();
			os_thread_destroy(&thread);
			break;
		}

		// End processing.
		_wait_cancel(fd);	// just in case
		_wait_for_done();
		os_thread_destroy(&thread);

		if (_thread_err[0])
		{
			errs++;
			printf("FAIL <---  (%d. Failure in wait thread: %s)\n", __LINE__, _thread_err);
			break;
		}

		if (_thread_wait.flags & GSC_WAIT_FLAG_TIMEOUT)
		{
			errs++;
			printf("FAIL <---  (%d. Wait for interrupt timed out.)\n", __LINE__);
			break;
		}

		if (_thread_wait.flags & GSC_WAIT_FLAG_CANCEL)
		{
			errs++;
			printf("FAIL <---  (%d. Wait for interrupt was cancelled.)\n", __LINE__);
			break;
		}

		if (_thread_wait.flags & ~GSC_WAIT_FLAG_DONE)
		{
			errs++;
			printf("FAIL <---  (%d. Invalid wait flags results: 0x%lX)\n", __LINE__, (long) _thread_wait.flags);
			break;
		}

		if (_thread_wait.flags & GSC_WAIT_FLAG_DONE)
		{
			printf("PASS\n");
			break;
		}

		errs++;
		printf("FAIL <---  (%d. Invalid wait flags results: 0x%lX)\n", __LINE__, (long) _thread_wait.flags);
		break;
	}

	os_sem_destroy(&_sem_enter);
	os_sem_destroy(&_sem_exit);
	return(errs);
}



//*****************************************************************************
int perform_tests(const args_t* args)
{
	int	errs	= 0;
	int	i;

	for (i = 0; _tests[i].name; i++)
		errs	+= _test_irq(args->fd, &_tests[i]);

	return(errs);
}


