// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/24DSI64C200K/24DSI64C200K_linux_1.x_gsc_dn/utils/util_fsamp_ai.c $
// $Rev: 52653 $
// $Date: 2023-03-21 10:02:20 -0500 (Tue, 21 Mar 2023) $

// 24DSI64C200K: Utilities: source file

#include "main.h"



// data types *****************************************************************

typedef struct
{
	s32		fsamp;			// The requested sample rate.
	double	rate;			// The computed sample rate.
	double	delta;			// The computed difference. Can be pos or neg.
	double	error;			// The computed difference. Always positive.

	s32		fref;

	s32		nvco_optimal;	// Use the narrower optimal Nvco range?
	s32		nvco_max;		// The maximum valid value.
	s32		nvco_min;		// The minimum valid value.
	s32		nvco_opt_max;	// The maximum optimal value.
	s32		nvco_opt_min;	// The minimum optimal value.
	s32		nvco;			// The numeric value to use for computations.

	s32		nref_optimal;	// Use the narrower optimal Nref range?
	s32		nref_max;		// The maximum valid value.
	s32		nref_min;		// The minimum valid value.
	s32		nref_opt_max;	// The maximum optimal value.
	s32		nref_opt_min;	// The minimum optimal value.
	s32		nref;			// The numeric value to use for computations.

	int		fgen_div_auto;	// Auto select the Fgen divisor?
	s32		fgen_div;		// The API option.
	s32		fgen_divisor;	// The numeric value to use for computations.
	s32		fgen_min;		// The minimum valid Fgen value.
	s32		fgen_max;		// The maximum valid Fgen value.
	s32		fgen;			// The value computed.

	int		mclk_div_auto;	// Auto select the Mclk divisor?
	s32		mclk_div;		// The API option.
	s32		mclk_divisor;	// The numeric value to use for computations.

	int		over_auto;		// Auto select the over sampling factor?
	s32		over_sample;	// The API option.
	s32		over_divisor;	// The numeric value to use for computations.

} fsamp_t;



//*****************************************************************************
static void _fsamp_compute(fsamp_t* data)
{
	double	delta;
	s32		denominator;
	double	error;
	s32		fgen;
	double	rate;

	// All values are known to be valid by the time we get here.

	fgen		= (s32) ((long long) data->fref * data->nvco / data->nref);
	denominator	= 2
				* data->fgen_divisor
				* data->mclk_divisor
				* data->over_divisor;
	rate		= (double) fgen / denominator;
	delta		= rate - data->fsamp;

	if (delta < 0)
		error	= -delta;
	else
		error	= delta;

	data->fgen	= fgen;
	data->rate	= rate;
	data->delta	= delta;
	data->error	= error;
}




//*****************************************************************************
static void _over_sample_compute(fsamp_t* data)
{
	typedef struct
	{
		s32	divisor;
		s32	over_sample;
	} list_t;

	static const list_t	primary[]	=
	{
		// divisor	over_sample
		{ 32,		DSI64C200K_OVER_SAMPLE_32	},
		{ 64,		DSI64C200K_OVER_SAMPLE_64	},
		{ 128,		DSI64C200K_OVER_SAMPLE_128	},
		{ 1024,		DSI64C200K_OVER_SAMPLE_1024	},
		{ -1,		-1							}
	};

	static list_t	secondary[]	=
	{
		// divisor	over_sample
		{ -1,		-1	},	// modified if the caller requests a specific option
		{ -1,		-1	}
	};

	int				i;
	const list_t*	list;
	fsamp_t			tmp;

	if (data->over_auto)
	{
		list	= primary;
	}
	else
	{
		list						= secondary;
		secondary[0].divisor		= data->over_divisor;
		secondary[0].over_sample	= data->over_sample;
	}

	for (i = 0; list[i].divisor > 0; i++)
	{
		tmp					= data[0];
		tmp.over_sample		= list[i].over_sample;
		tmp.over_divisor	= list[i].divisor;
		tmp.error			= 999999999;
		_fsamp_compute(&tmp);

		if (data->error > tmp.error)
			data[0]	= tmp;
	}
}



//*****************************************************************************
static int _over_sample_decode(s32 over_sample, fsamp_t* data)
{
	int	errs	= 0;

	data->over_auto		= 0;
	data->over_sample	= over_sample;

	switch (over_sample)
	{
		default:

			errs	= 1;
			printf("FAIL <---  (Unrecognized Over Sample option: %ld)\n", (long) over_sample);
			break;

		case -1:							data->over_auto		= 1;	break;
		case DSI64C200K_OVER_SAMPLE_32:		data->over_divisor	= 32;	break;
		case DSI64C200K_OVER_SAMPLE_64:		data->over_divisor	= 64;	break;
		case DSI64C200K_OVER_SAMPLE_128:	data->over_divisor	= 128;	break;
		case DSI64C200K_OVER_SAMPLE_1024:	data->over_divisor	= 1024;	break;
	}

	return(errs);
}



//*****************************************************************************
static void _mclk_div_compute(fsamp_t* data)
{
	typedef struct
	{
		s32	divisor;
		s32	mclk_div;
	} list_t;

	static const list_t	primary[]	=
	{
		// divisor	mclk_div
		{ 4,		DSI64C200K_MCLK_DIV_4_FAST		},
		{ 8,		DSI64C200K_MCLK_DIV_8_MEDIAN	},
		{ 32,		DSI64C200K_MCLK_DIV_32_ECO		},
		{ -1,		-1								}
	};

	static list_t	secondary[]	=
	{
		// divisor	mclk_div
		{ -1,		-1	},	// modified if the caller requests a specific option
		{ -1,		-1	}
	};

	int				i;
	const list_t*	list;
	fsamp_t			tmp;

	if (data->mclk_div_auto)
	{
		list	= primary;
	}
	else
	{
		list					= secondary;
		secondary[0].divisor	= data->mclk_divisor;
		secondary[0].mclk_div	= data->mclk_div;
	}

	for (i = 0; list[i].divisor > 0; i++)
	{
		tmp					= data[0];
		tmp.mclk_div		= list[i].mclk_div;
		tmp.mclk_divisor	= list[i].divisor;
		tmp.error			= 999999999;
		_over_sample_compute(&tmp);

		if (data->error > tmp.error)
			data[0]	= tmp;
	}
}



//*****************************************************************************
static int _mclk_div_decode(s32 mclk_div, fsamp_t* data)
{
	int	errs	= 0;

	data->mclk_div_auto	= 0;
	data->mclk_div		= mclk_div;

	switch (mclk_div)
	{
		default:

			errs	= 1;
			printf("FAIL <---  (Unrecognized Mclk Div option: %ld)\n", (long) mclk_div);
			break;

		case -1:							data->mclk_div_auto	= 1;	break;
		case DSI64C200K_MCLK_DIV_4_FAST:	data->mclk_divisor	= 4;	break;
		case DSI64C200K_MCLK_DIV_8_MEDIAN:	data->mclk_divisor	= 8;	break;
		case DSI64C200K_MCLK_DIV_32_ECO:	data->mclk_divisor	= 32;	break;
	}

	return(errs);
}



//*****************************************************************************
static void _fgen_div_compute(fsamp_t* data)
{
	typedef struct
	{
		s32	divisor;
		s32	fgen_div;
	} list_t;

	static const list_t	primary[]	=
	{
		// divisor	fgen_div
		{ 1,		DSI64C200K_FGEN_DIV_1	},
		{ 2,		DSI64C200K_FGEN_DIV_2	},
		{ 4,		DSI64C200K_FGEN_DIV_4	},
		{ 8,		DSI64C200K_FGEN_DIV_8	},
		{ 16,		DSI64C200K_FGEN_DIV_16	},
		{ -1,		-1						}
	};

	static list_t	secondary[]	=
	{
		// divisor	fgen_div
		{ -1,		-1	},	// modified if the caller requests a specific option
		{ -1,		-1	}
	};

	int				i;
	const list_t*	list;
	fsamp_t			tmp;


	if (data->fgen_div_auto)
	{
		list	= primary;
	}
	else
	{
		list					= secondary;
		secondary[0].divisor	= data->fgen_divisor;
		secondary[0].fgen_div	= data->fgen_div;
	}

	for (i = 0; list[i].divisor > 0; i++)
	{
		tmp					= data[0];
		tmp.fgen_div		= list[i].fgen_div;
		tmp.fgen_divisor	= list[i].divisor;
		tmp.error			= 999999999;
		_mclk_div_compute(&tmp);

		if (data->error > tmp.error)
			data[0]	= tmp;
	}
}



//*****************************************************************************
static int _fgen_div_decode(s32 fgen_div, fsamp_t* data)
{
	int	errs	= 0;

	data->fgen_div_auto	= 0;
	data->fgen_div		= fgen_div;

	switch (fgen_div)
	{
		default:

			errs	= 1;
			printf("FAIL <---  (Unrecognized Fgen Div option: %ld)\n", (long) fgen_div);
			break;

		case -1:						data->fgen_div_auto	= 1;	break;
		case DSI64C200K_FGEN_DIV_1:		data->fgen_divisor	= 1;	break;
		case DSI64C200K_FGEN_DIV_2:		data->fgen_divisor	= 2;	break;
		case DSI64C200K_FGEN_DIV_4:		data->fgen_divisor	= 4;	break;
		case DSI64C200K_FGEN_DIV_8:		data->fgen_divisor	= 8;	break;
		case DSI64C200K_FGEN_DIV_16:	data->fgen_divisor	= 16;	break;
	}

	return(errs);
}



//*****************************************************************************
static void _fgen_compute(fsamp_t* data)
{
	s32		fgen;
	fsamp_t	tmp;

	fgen	= (s32) ((long long) data->fref * data->nvco / data->nref);

	if ((fgen >= data->fgen_min) && (fgen <= data->fgen_max))
	{
		tmp			= data[0];
		tmp.fgen	= fgen;
		tmp.error	= 999999999;
		_fgen_div_compute(&tmp);

		if (data->error > tmp.error)
			data[0]	= tmp;
	}
}



//*****************************************************************************
static void _nref_compute(fsamp_t* data)
{
	int		max;
	int		min;
	int		nref;
	double	ratio;
	fsamp_t	tmp;

	if (data->nref_optimal)
	{
		max	= data->nref_opt_max;
		min	= data->nref_opt_min;
	}
	else
	{
		max	= data->nref_max;
		min	= data->nref_min;
	}

	for (nref = min; nref <= max; nref++)
	{
		ratio	= (double) data->nvco / nref;

		if ((ratio < 1.04) || (ratio > 2.08))
			continue;

		tmp			= data[0];
		tmp.nref	= nref;
		tmp.error	= 999999999;
		_fgen_compute(&tmp);

		if (data->error > tmp.error)
			data[0]	= tmp;
	}
}



//*****************************************************************************
static void _nvco_compute(fsamp_t* data)
{
	int		max;
	int		min;
	int		nvco;
	fsamp_t	tmp;

	if (data->nvco_optimal)
	{
		max	= data->nvco_opt_max;
		min	= data->nvco_opt_min;
	}
	else
	{
		max	= data->nvco_max;
		min	= data->nvco_min;
	}

	for (nvco = min; nvco <= max; nvco++)
	{
		tmp			= data[0];
		tmp.nvco	= nvco;
		tmp.error	= 999999999;
		_nref_compute(&tmp);

		if (data->error > tmp.error)
			data[0]	= tmp;
	}
}



/******************************************************************************
*
*	Function:	dsi64c200k_fsamp_ai_compute
*
*	Purpose:
*
*		Calculate the best values for the specified sample rate.
*
*	Arguments:
*
*		fd			The handle for the device to access.
*
*		index		The index of the device to access. Ignore if < 0.
*
*		verbose		Work verbosely.
*
*		fref		The is the Fref value.
*
*		fsamp		The desired sample rate. Updated before return to the
*					achieved rate.
*
*		optimal		Are the optimal Nvco and Nref ranges to be used?
*
*		nvco		Put the resulting Nvco value here.
*
*		nref		Put the resulting Nref value here.
*
*		fgen_div	Must be initialized to -1 for auto-select, or to a valid
*					API option.
*
*		mclk_div	Must be initialized to -1 for auto-select, or to a valid
*					API option.
*
*		over_sample	Must be initialized to -1 for auto-select, or to a valid
*					API option.
*
*		rate		The achieved floating point Fsamp is reported here.
*
*	Returned:
*
*		>= 0	The number of errors encounterred.
*
******************************************************************************/

int dsi64c200k_fsamp_ai_compute(
	int		fd,
	int		index,
	int		verbose,
	s32		fref,
	s32		fsamp,
	int		optimal,
	s32*	nvco,
	s32*	nref,
	s32*	fgen_div,
	s32*	mclk_div,
	s32*	over_sample,
	double*	rate)
{
	fsamp_t	data;
	int		errs	= 0;

	for (;;)	// A convenience loop.
	{
		if (verbose)
		{
			gsc_label_index("AI Sample Rate", index);
			gsc_label_long_comma(fsamp);
			printf(" S/S");
		}

		if ((nvco == NULL) || (nref == NULL) ||
			(fgen_div == NULL) || (mclk_div == NULL) ||
			(over_sample == NULL) || (rate == NULL))
		{
			errs++;
			printf("FAIL <---  (dsi64c200k_fsamp_ai_compute: NULL argument(s))\n");
			break;
		}

		// Initialization and setup.
		memset(&data, 0, sizeof(data));
		data.fref			= fref;
		data.fsamp			= fsamp;
		data.error			= 0x7FFFFFFF;
		data.nvco_optimal	= optimal ? 1 : 0;
		data.nref_optimal	= optimal ? 1 : 0;

		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_NVCO_MAX, &data.nvco_max);
		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_NVCO_MIN, &data.nvco_min);
		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_NVCO_OPT_MAX, &data.nvco_opt_max);
		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_NVCO_OPT_MIN, &data.nvco_opt_min);

		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_NREF_MAX, &data.nref_max);
		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_NREF_MIN, &data.nref_min);
		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_NREF_OPT_MAX, &data.nref_opt_max);
		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_NREF_OPT_MIN, &data.nref_opt_min);

		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_FGEN_MAX, &data.fgen_max);
		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_FGEN_MIN, &data.fgen_min);
		errs	+= _fgen_div_decode(fgen_div[0], &data);

		errs	+= _mclk_div_decode(mclk_div[0], &data);

		errs	+= _over_sample_decode(over_sample[0], &data);

		// Do the work to find the best configuration.

		if (errs == 0)
		{
			_nvco_compute(&data);

			nvco[0]			= data.nvco;
			nref[0]			= data.nref;
			fgen_div[0]		= data.fgen_div;
			mclk_div[0]		= data.mclk_div;
			over_sample[0]	= data.over_sample;
			rate[0]			= data.rate;
		}

		// Report the results.

		if (verbose == 0)
		{
		}
		else if (errs)
		{
			printf("FAIL <---\n");
		}
		else
		{
			printf("  (Fref ");
			gsc_label_long_comma(data.fref);
			printf(", Nvco ");
			gsc_label_long_comma(data.nvco);
			printf(", Nref ");
			gsc_label_long_comma(data.nref);
			printf(", Fgen Div ");
			gsc_label_long_comma(data.fgen_divisor);
			printf(", Mclk Div ");
			gsc_label_long_comma(data.mclk_divisor);
			printf(", OSF ");
			gsc_label_long_comma(data.over_divisor);
			printf(", Rate %.3lf)\n", (double) data.rate);
		}

		break;
	}

	return(errs);
}



/******************************************************************************
*
*	Function:	dsi64c200k_fsamp_ai_report
*
*	Purpose:
*
*		Determine and report the sample rate for the specified channel.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		chan	The index of the channel of interest.
*
*		sps		Store the sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int dsi64c200k_fsamp_ai_report(int fd, int index, int verbose, int chan, s32* sps)
{
	s32		ai_buf;
	char	buf[64];
	s32		burst;
	s32		clk_src;
	fsamp_t	data;
	int		errs		= 0;
	s32		fsamp		= 0;
	s32		grp_active;
	s32		grp_bit;
	s32		grp_no;
	s32		max;
	s32		qty;
	s32		ready;

	for (;;)	// A convenience loop.
	{
		if (verbose)
		{
			sprintf(buf, "Channel %d", chan);
			gsc_label_index(buf, index);
		}

		memset(&data, 0, sizeof(data));

		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_CHANNEL_MAX, &max);
		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_CHANNEL_QTY, &qty);
		errs	+= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_FREF_DEFAULT, &data.fref);
		errs	+= dsi64c200k_ai_buf_enable		(fd, -1, 0, -1, &ai_buf);
		errs	+= dsi64c200k_burst_enable		(fd, -1, 0, -1, &burst);
		errs	+= dsi64c200k_chan_grp_active	(fd, -1, 0, -1, &grp_active);
		errs	+= dsi64c200k_channels_ready	(fd, -1, 0, DSI64C200K_CHANNELS_READY_WAIT, &ready);
		errs	+= dsi64c200k_clk_src			(fd, -1, 0, -1, &clk_src);
		errs	+= dsi64c200k_nvco				(fd, -1, 0, -1, &data.nvco);
		errs	+= dsi64c200k_nref				(fd, -1, 0, -1, &data.nref);
		errs	+= dsi64c200k_fgen_div			(fd, -1, 0, -1, &data.fgen_div);
		errs	+= dsi64c200k_mclk_div			(fd, -1, 0, -1, &data.mclk_div);
		errs	+= dsi64c200k_over_sample		(fd, -1, 0, -1, &data.over_sample);

		errs	+= _fgen_div_decode(data.fgen_div, &data);
		errs	+= _mclk_div_decode(data.mclk_div, &data);
		errs	+= _over_sample_decode(data.over_sample, &data);

		if (errs)
			break;

		if (chan >= max)
		{
			if (verbose)
				printf("0 S/S  (Channel %d an invalid channel index)\n", chan);

			break;
		}

		if (chan >= qty)
		{
			if (verbose)
				printf("0 S/S  (Channel %d not installed)\n", chan);

			break;
		}

		grp_no	= chan / 8;
		grp_bit	= 0x1 << grp_no;

		if ((grp_active & grp_bit) == 0)
		{
			if (verbose)
				printf("0 S/S  (Channel Group %d disabled)\n", grp_no);

			break;
		}

		if (ai_buf == DSI64C200K_AI_BUF_ENABLE_NO)
		{
			if (verbose)
				printf("0 S/S  (Input Buffer disabled)\n");

			break;
		}

		if (ready == DSI64C200K_CHANNELS_READY_NO)
		{
			if (verbose)
				printf("0 S/S  (Channel Ready negated)\n");

			break;
		}

		if (burst == DSI64C200K_BURST_ENABLE_YES)
		{
			if (verbose)
				printf("0 S/S  (Rate unknown, Bursting enabled)\n");

			break;
		}

		if (clk_src != DSI64C200K_CLK_SRC_RATE_GEN)
		{
			if (verbose)
				printf("0 S/S  (Rate unknown, External clocking selected)\n");

			break;
		}

		_fsamp_compute(&data);
		fsamp	= (s32) (data.rate + 0.5);

		if (verbose)
		{
			gsc_label_long_comma(fsamp);
			printf(" S/S  (Fref ");
			gsc_label_long_comma(data.fref);
			printf(", Nvco ");
			gsc_label_long_comma(data.nvco);
			printf(", Nref ");
			gsc_label_long_comma(data.nref);
			printf(", Fgen Div ");
			gsc_label_long_comma(data.fgen_divisor);
			printf(", Mclk Div ");
			gsc_label_long_comma(data.mclk_divisor);
			printf(", OSF ");
			gsc_label_long_comma(data.over_divisor);
			printf(", Rate %.3lf)\n", (double) data.rate);
		}

		break;
	}


	if (sps)
		sps[0]	= fsamp;

	return(errs);
}



/******************************************************************************
*
*	Function:	dsi64c200k_fsamp_ai_report_all
*
*	Purpose:
*
*		Determine and report the sample rate for all channels.
*
*	Arguments:
*
*		fd		The handle to use to access the driver.
*
*		index	The index of the device to access. Ignore if < 0.
*
*		verbose	Work verbosely?
*
*		sps		Report the overall sample rate here, if non-NULL.
*
*	Returned:
*
*		>= 0	The number of errors encountered here.
*
******************************************************************************/

int dsi64c200k_fsamp_ai_report_all(int fd, int index, int verbose, s32* sps)
{
	s32	chans;
	int	chan;
	int	errs;
	s32	fsamp;
	s32	total	= 0;

	if (verbose)
		gsc_label_index("Sample Rates", index);

	errs	= dsi64c200k_query(fd, -1, 0, DSI64C200K_QUERY_CHANNEL_QTY, &chans);


	if (errs == 0)
	{
		if (verbose)
			printf("\n");

		gsc_label_level_inc();

		for (chan = 0; chan < chans; chan++)
		{
			errs	+= dsi64c200k_fsamp_ai_report(fd, -1, verbose, chan, &fsamp);
			total	+= fsamp;
		}

		if (verbose)
		{
			gsc_label("Overall Rate");
			gsc_label_long_comma(total);
			printf(" S/S\n");
		}

		gsc_label_level_dec();
	}

	if (sps)
		sps[0]	= total;

	return(errs);
}


